#!/usr/bin/perl

use strict;
use warnings;
use POSIX;

##################################################
#-------------------------------------------------
#-----------Viterbi-decoding HTK------------------
#-------------------------------------------------
##################################################

# Choose number of mixtures and penalty for jumping between class models
my $mix = 32; my $wmix = "32";
my $wordpen = -80; my $wp = "pm_80";
 
 # Assuming gmm models are trained and placed in ... :
my @modeldirs = ("models1", "models2", "models3", "models4", "models5");

# Using the following configure file
my $cnfg = "mhj_20_mfc.cfg";

# and the following test sets :
my @datasets= ("lists/test1.scp",  "lists/test2.scp", "lists/test3.scp", "lists/test4.scp", "lists/test5.scp" );

# Class_list, dictionary and network must exist
my $cl_list = "class_pause_ny.list";
my $dict = "birds_ny.dict";
my $network = "birds_ny.net";

my $labmlf = "Kvirrevitt_ny.mlf";

# Establish folder to put rec-files
my $resultDir = "Results_hmm";
unless(-d "$resultDir") {system("mkdir $resultDir");}
my $resultfile = $resultDir."/dummy.res";
system "rm $resultfile ";
system "touch $resultfile";

# Loop over the five training/test-set partitions of the data.
my $set = -1;
foreach my $modeldir (@modeldirs){

	$set = $set +1;
    my $setind = $set +1;
	my $testset = $datasets[$set];
	$modeldir = $modeldirs[$set];
    my $mmf = "$modeldir/state1_$mix.mmf";
	print "speakerlist : $testset\n";
	# printf "Results are placed in $recmlf and $resultfile\n";
    my $recmlf = $resultDir."/test".$setind."_".$wmix."_".$wp.".rec";
    

	system "HVite -o N -C $cnfg  -A -t 0.0 -s 0 -p $wordpen -H $mmf -y rec -S $testset -i $recmlf -w $network $dict $cl_list >>  $resultfile";
    system "HResults -p  -A -I $labmlf $cl_list $recmlf >> $resultfile";
}




##################################################
#-------------------------------------------------
#-------------Histogram program-------------------
#-------------------------------------------------
##################################################

my $pm = 80;
my $mix = 32;
my $set = 4;
my $type = 'test';
my $modelName;
my @modelList = ();
my $Nclass = 0;
my $classfile = 'class_pause_ny.list'; # list of all legal classes

open(LIST,"$classfile") || die "Could not open $classfile for reading\n";
while(<LIST>){
	chomp;
	$modelName = $_;
	push (@modelList,$modelName);
	$Nclass = $Nclass + 1;
}
close(LIST);

	
my $Nclass_birds = $Nclass - 1; # pause is not a valid class

my $Best_1 = 0;
my $Best_2 = 0;
my $Best_3 = 0;

for ($set = 1; $set <= 5; $set++){ #all five sets

my $recfile = "Results_hmm/test$set\_$mix\_pm\_$pm.rec";

open (REC,"$recfile") || die "Could not open $recfile";

my $line1 = <REC>; # first line, MLF line
my $linelen;
my $seglen;
my @sent = ();
my $start;
my $end;
my $label;
my $counter2 = -1;
my $counter1 = -1;
#$counter1 = -1;
my $tars2;
my $sind2;
my $sind_old2;
my $title;
my @trec;
my $scale = (20 * (10**4));
my @frame_distr;
my @file_class;
my @num_hit = @file_class;
my $length_trec = 0;
my $jnum;

while ($line1 = <REC>){
	chomp($line1);
	$linelen = length($line1);
	if ($line1 =~ /\.rec\"/) 
	{
		$sind_old2 = 0;
		@trec = ();
		$length_trec = 0;
		$title = $line1;
		$counter1 = $counter1 + 1;   # count number of files
		for (my $i = 0; $i < $Nclass; $i++){
			$frame_distr[$i][$counter1] = 0;
			my $str = $title;
			my $substr = $modelList[$i];
			if (index($str, $substr) != -1) {
				$counter2 = $counter2 + 1;
				$file_class[$counter1] = $i;
				$num_hit[$counter1] += 1;
			}
		} 
	}
	elsif ($linelen > 4) 
	{ 
		chomp($line1); 
		($start, $end, $label, ) = split(/ /,$line1);  # extracting info from rec-file
		for (my $i = 0; $i < $Nclass; $i++){
			if ($label eq $modelList[$i]){
				$tars2 = $i;  #comparing against list of birds
			}
		}
		$sind2 = floor($end / $scale);  # end of file
		$seglen = $sind2 - $sind_old2; #length of each recognized bird
		#$sind_old2 = $sind2;
		my @ones = (1) x $seglen;
		foreach (@ones){
			$_ = $_ * $tars2;
			$length_trec = $length_trec + 1;
		}
		push (@trec, @ones);
		for (my $j = 0; $j < $Nclass_birds; $j++){
			my $counter_trec = 0;
			for (my $i = $sind_old2; $i < $length_trec; $i++){
				if ($trec[$i] == $j){
					$counter_trec = $counter_trec + 1;
				}
			}
			$frame_distr[$j][$counter1] += $counter_trec;
		}
		$sind_old2 = $sind2;
	}	
}

close (REC);

#------------------------------------------
#-------Normalized Histogram---------------
#------------------------------------------

my @frame_distr_sum;
my @frame_distr_norm;


for (my $i = 0; $i <= $counter1; $i++){
	$frame_distr_sum[$i] = 0;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$frame_distr_sum[$i] = $frame_distr_sum[$i] + $frame_distr[$j][$i];
	}
}

my $sum;

for (my $i = 0; $i <= $counter1; $i++){
	$sum = 0;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$frame_distr_norm[$j][$i] = $frame_distr[$j][$i] / $frame_distr_sum[$i];
		$sum = $sum + $frame_distr_norm[$j][$i];
	}
}

############################################
#-------------------------------------------
#------------Classification-----------------
#-------------------------------------------
############################################

#---------------------------------------------------------
#----------------------Importing relevant data------------
#---------------------------------------------------------

for (my $i = 0; $i <= $counter1; $i++){
	$frame_distr_norm[$Nclass_birds][$i] = 1;
}

my @Ttest = ();

for (my $i = 0; $i <= $counter1; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$Ttest[$j][$i] = 0;
		if ($file_class[$i] == $j){
			$Ttest[$j][$i] = $Ttest[$j][$i] + 1;
		}
	}
}

#----------------------------------------
#----Subroutine matrix multiplication----
#----------------------------------------

sub matrix_multiply {
    my ($r_mat1,$r_mat2)=@_;
    my ($r_product);
    my ($r1,$c1)=matrix_count_rows_cols($r_mat1);
    my ($r2,$c2)=matrix_count_rows_cols($r_mat2);

    
    die "matrix 1 has $c1 columns and matrix 2 has $r2 rows>" 
        . " Cannot multiply\n" unless ($c1==$r2);
    for (my $i=0;$i<$r1;$i++) {
        for (my $j=0;$j<$c2;$j++) {
            my $sum=0;
            for (my $k=0;$k<$c1;$k++) {
                $sum+=$r_mat1->[$i][$k]*$r_mat2->[$k][$j];
            }
            $r_product->[$i][$j]=$sum;
        }
    }
    $r_product;
}

sub matrix_count_rows_cols { 
    my ($r_mat)=@_;
    my $num_rows=@$r_mat;
    my $num_cols=@{$r_mat->[0]};
    ($num_rows,$num_cols);
}

#--------------------------------------
#----------Defining w matrix-----------
#--------------------------------------

my @w;
my $line;
my $i = 0;
my $textfile = "classificationMatrix$set.txt";

open (TEXT,"$textfile") || die "Could not open $textfile";

while ($line = <TEXT>){
	chomp($line);
	
	(my $class1, my $class2, my $class3, my $class4, my $class5, my $class6, my $class7, my $class8, my $class9, my$class10, my $class11, my $class12, my $class13, my $class14, my $class15, my $class16, my $class17, my $class18, my $class19, my $class20, my $class21, my $class22 ) = split(/ /,$line);
	
	
	#print "\n";
	
		$w[$i][0] = $class1;
		$w[$i][1] = $class2;
		$w[$i][2] = $class3;
		$w[$i][3] = $class4;
		$w[$i][4] = $class5;
		$w[$i][5] = $class6;
		$w[$i][6] = $class7;
		$w[$i][7] = $class8;
		$w[$i][8] = $class9;
		$w[$i][9] = $class10;
		$w[$i][10] = $class11;
		$w[$i][11] = $class12;
		$w[$i][12] = $class13;
		$w[$i][13] = $class14;
		$w[$i][14] = $class15;
		$w[$i][15] = $class16;
		$w[$i][16] = $class17;
		$w[$i][17] = $class18;
		$w[$i][18] = $class19;
		$w[$i][19] = $class20;
		$w[$i][20] = $class21;
		$w[$i][21] = $class22;
		
		$i = $i +1;
	
}

close(TEXT);


#------------------------------------------------
#-----Initializing misclassification matrix------
#------------------------------------------------

my @ctest1;
my @ctest2;
my @ctest3;

for (my $i = 0; $i < $Nclass_birds; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$ctest1[$i][$j] = 0;
		$ctest2[$i][$j] = 0;
		$ctest3[$i][$j] = 0;
	}
}


#-----------------------------------------------
#---------Forward calculations------------------
#-----------------------------------------------

my @zt;
my @bt;

@zt = @{ matrix_multiply(\@w, \@frame_distr_norm) };

for (my $i = 0; $i <= $counter1; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$bt[$j][$i] = exp($zt[$j][$i]);
	}
}

my @bt_sum;
	
for (my $i = 0; $i <= $counter1; $i++){
	$bt_sum[$i] = 0;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$bt_sum[$i] = $bt_sum[$i] + $bt[$j][$i];
	}
}
	
my @sum;
my @bt_norm;

for (my $i = 0; $i <= $counter1; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$bt_norm[$j][$i] = $bt[$j][$i] / $bt_sum[$i];
	}
}


#--------------------------------------------------
#------Calculating misclassification matrices------
#--------------------------------------------------

#####Sorting bd matrix
my @bt_norm_sorted;
my @bt_norm_sorted_temp;


for (my $i = 0; $i <= $counter1; $i++){
	@bt_norm_sorted_temp = sort {$b->[$i] <=> $a->[$i]} @bt_norm;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$bt_norm_sorted[$j][$i] = $bt_norm_sorted_temp[$j][$i];
	}
}
	
######Finding indexes of the sorted matrix
my @file_rec;

for (my $i = 0; $i <= $counter1; $i++){
	$file_rec[0][$i] = -1;
	$file_rec[1][$i] = -1;
	$file_rec[2][$i] = -1;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		if ($bt_norm_sorted[0][$i] == $bt_norm[$j][$i] && $bt_norm_sorted[0][$i] > 0.01){   #If the score is low, do not show all 3best				
			$file_rec[0][$i] = $j;
		}
		if ($bt_norm_sorted[1][$i] == $bt_norm[$j][$i] && $bt_norm_sorted[1][$i] > 0.01){	
			$file_rec[1][$i] = $j;
		}
		if ($bt_norm_sorted[2][$i] == $bt_norm[$j][$i] && $bt_norm_sorted[2][$i] > 0.01){				
			$file_rec[2][$i] = $j;
		} 
	}
}

#######Display wich bird is recognized
my @recognized;

for (my $i = 0; $i <= $counter1; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		for (my $k = 0; $k < 3; $k++){
			$recognized[$k][$i] = $file_rec[$k][$i];
			if ($recognized[$k][$i] == -1){
				$recognized[$k][$i] = "No bird recognized";
			}
			else {
				$recognized[$k][$i] = "$modelList[$recognized[$k][$i]]";
			}
		}
	}
}		                            
	
###########Calculating misclassification matices	
my @classtest;
my @cdev;
	
for (my $i = 0; $i <= $counter1; $i++){
	$classtest[0][$i] = $file_rec[0][$i];
	$classtest[1][$i] = $file_rec[1][$i];
	$classtest[2][$i] = $file_rec[2][$i];
}
	
my $count_file_class1 = 0;
my $count_file_class2 = 0;
my $count_file_class3 = 0;
	
	
##Mis matrix
	
for (my $i = 0; $i <= $counter1; $i++){
	$count_file_class1 = 0;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		for (my $k = 0; $k < $Nclass_birds; $k++){
			if ($file_class[$i] == $j && $file_rec[0][$i] == $k){
				$count_file_class1 +=1;
				$ctest1[$j][$k] = $ctest1[$j][$k] + $count_file_class1; 
			}
		}
	}
}
	

####Number of correct classifications
	
$count_file_class1 = 0;
$count_file_class2 = 0;
$count_file_class3 = 0;
	
for (my $i = 0; $i <= $counter1; $i++){
	if ($file_rec[0][$i] == $file_class[$i]){
		$count_file_class1 +=1;
	}
	if ($file_rec[1][$i] == $file_class[$i]){
		$count_file_class2 +=1;
	}
	if ($file_rec[2][$i] == $file_class[$i]){
		$count_file_class3 +=1;
	}
}
	
my $numcorr1d = $count_file_class1;
my $numcorr2d = $count_file_class1 + $count_file_class2;
my $numcorr3d = $count_file_class1 + $count_file_class2 + $count_file_class3; 
	
my @perr1;
my @perr2;
my @perr3;
my @dzd;

for (my $i = 0; $i < 3; $i++){
	if ($count_file_class1 != 0){				
		my $score1 = 100 * (1 - ($numcorr1d / ($counter1 + 1)));
		$perr1[$set] = "$score1";
	}
	if ($count_file_class2 != 0){
		my $score2 = 100 * (1 - ($numcorr2d / ($counter1 + 1)));
		$perr2[$set] = "$score2";
	}
	if ($count_file_class3 != 0){
		my $score3 = 100 * (1 - ($numcorr3d / ($counter1 + 1)));
		$perr3[$set] = "$score3";
	}
	
}	

print "\n";
print "\n";		
	
print "$perr1[$set], $perr2[$set], $perr3[$set]\n \n";

$Best_1 = $Best_1 + $perr1[$set]/5;
$Best_2 = $Best_2 + $perr2[$set]/5;
$Best_3 = $Best_3 + $perr3[$set]/5;

print "\n";
print "\n";


open(my $fh, '>', 'adap_recognized.list');
for (my $i = 0; $i <= $counter1; $i++){
	print $fh $recognized[0][$i];
	print $fh "\n";
}

}

print "Average error rate: \n";
print "$Best_1    $Best_2     $Best_3 \n";  #average  error
