#!/usr/bin/perl

use strict;
use warnings;
use POSIX;

##################################################
#-------------------------------------------------
#-------------Histogram program-------------------
#-------------------------------------------------
##################################################

my $pm = 80;
my $mix = 32;
my $set = 2;
my $type = 'train';
my $modelName;
my @modelList = ();
my $Nclass = 0;

my $classfile = 'class_pause_ny.list'; # list of all legal classes

open(LIST,"$classfile") || die "Could not open $classfile for reading\n";
while(<LIST>){
	chomp;
	$modelName = $_;
	push (@modelList,$modelName);
	$Nclass = $Nclass + 1;
}
close(LIST);

	
my $Nclass_birds = $Nclass - 1; # pause is not a valid class

my $Best_1 = 0;
my $Best_2 = 0;
my $Best_3 = 0;

my @perr1;
my @perr2;
my @perr3;

for ($set = 1; $set <= 5; $set++){ #all five sets

my $recfile = "Results_hmm/train$set\_$mix\_pm\_$pm.rec";

my $linelen;
my $seglen;
my @sent = ();
my $start;
my $end;
my $label;
my $counter2 = -1;
my $counter1 = -1;
my $tars2;
my $sind2;
my $sind_old2;
my $title;
my @trec;
my $scale = (20 * (10**4));
my @frame_distr;
my @file_class;
my @num_hit = @file_class;
my $length_trec = 0;
my $jnum;

#-------------------------------------------------
#----------------Reading MLF file-----------------
#-------------------------------------------------


open (REC,"$recfile") || die "Could not open $recfile";

my $line1 = <REC>; # first line, MLF line

while ($line1 = <REC>){
	chomp($line1);
	$linelen = length($line1);
	if ($line1 =~ /\.rec\"/)     #New file
	{
		$sind_old2 = 0;
		@trec = ();
		$length_trec = 0;
		$title = $line1;
		$counter1 = $counter1 + 1;   # count number of files
		for (my $i = 0; $i < $Nclass; $i++){
			$frame_distr[$i][$counter1] = 0;
			my $str = $title;
			my $substr = $modelList[$i];
			if (index($str, $substr) != -1) {
				$counter2 = $counter2 + 1;
				$file_class[$counter1] = $i;
				$num_hit[$counter1] += 1;
			}
		} 
	}
	elsif ($linelen > 4) 
	{ 
		chomp($line1); 
		($start, $end, $label, ) = split(/ /,$line1);  # extracting info from rec-file
		for (my $i = 0; $i < $Nclass; $i++){
			if ($label eq $modelList[$i]){
				$tars2 = $i;  #comparing against list of birds
			}
		}
		$sind2 = floor($end / $scale);  # end of file
		$seglen = $sind2 - $sind_old2; #length of each recognized bird
		#$sind_old2 = $sind2;
		my @ones = (1) x $seglen;
		foreach (@ones){
			$_ = $_ * $tars2;
			$length_trec = $length_trec + 1;
		}
		push (@trec, @ones);
		for (my $j = 0; $j < $Nclass_birds; $j++){
			my $counter_trec = 0;
			for (my $i = $sind_old2; $i < $length_trec; $i++){
				if ($trec[$i] == $j){
					$counter_trec = $counter_trec + 1;
				}
			}
			$frame_distr[$j][$counter1] += $counter_trec;
		}
		$sind_old2 = $sind2;
	}	
}

close (REC);

#-------------------------------------------------------------------
#----------------------Normalized Histogram-------------------------
#-------------------------------------------------------------------

my @frame_distr_sum;
my @frame_distr_norm;


for (my $i = 0; $i <= $counter1; $i++){
	$frame_distr_sum[$i] = 0;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$frame_distr_sum[$i] = $frame_distr_sum[$i] + $frame_distr[$j][$i];
	}
}

my $sum;

for (my $i = 0; $i <= $counter1; $i++){
	$sum = 0;
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$frame_distr_norm[$j][$i] = $frame_distr[$j][$i] / $frame_distr_sum[$i];
		$sum = $sum + $frame_distr_norm[$j][$i];
	}
}

##################################################
#-------------------------------------------------
#----------------Classification-------------------
#-------------------------------------------------
##################################################

#-----------------------------------------------------------
#---------------Importing relevant data---------------------
#-----------------------------------------------------------

for (my $i = 0; $i <= $counter1; $i++){
	$frame_distr_norm[$Nclass_birds][$i] = 1;
}

my @Ttrain = ();

for (my $i = 0; $i <= $counter1; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		$Ttrain[$j][$i] = 0;
		if ($file_class[$i] == $j){
			$Ttrain[$j][$i] = $Ttrain[$j][$i] + 1;
		}
	}
}

#-------------------------------------------------------------------
#-----------------Subroutine matrix multiplication------------------
#-------------------------------------------------------------------

sub matrix_multiply {
    my ($r_mat1,$r_mat2)=@_;
    my ($r_product);
    my ($r1,$c1)=matrix_count_rows_cols($r_mat1);
    my ($r2,$c2)=matrix_count_rows_cols($r_mat2);
    
    die "matrix 1 has $c1 columns and matrix 2 has $r2 rows>" 
        . " Cannot multiply\n" unless ($c1==$r2);
    for (my $i=0;$i<$r1;$i++) {
        for (my $j=0;$j<$c2;$j++) {
            my $sum=0;
            for (my $k=0;$k<$c1;$k++) {
                $sum+=$r_mat1->[$i][$k]*$r_mat2->[$k][$j];
            }
            $r_product->[$i][$j]=$sum;
        }
    }
    $r_product;
}

sub matrix_count_rows_cols { 
    my ($r_mat)=@_;
    my $num_rows=@$r_mat;
    my $num_cols=@{$r_mat->[0]};
    ($num_rows,$num_cols);
}

#--------------------------------------------------
#-----------Subroutine transpose matrix------------
#--------------------------------------------------

sub pivot {
    my @src = @_;

    my $max_col = 0;
    $max_col < $#$_ and $max_col = $#$_ for @src;

    my @dest;
    for my $col (0..$max_col) {
        my @new_row;
        for my $row (0..$#src) {
            push @new_row, $src[$row][$col] // '';
        }
        push @dest, \@new_row;
    }

    return @dest;
}

#-----------------------------------------
#----------Subroutine sign----------------
#-----------------------------------------

sub sign {
    if ($_[0]*$_[1] > 0){
        return 1;
    }
    elsif ($_[0]*$_[1] < 0){
        return -1;
    }
    else{
        return 0;
    }
};

#---------------------------------
#------Training starts------------
#---------------------------------

my $totiter = 1000;

#--------------------------------------
#----------Defining w matrix-----------
#--------------------------------------

my @w;
my $line;
my $i = 0;
my $textfile = "w.txt";  #Initially using the same w matrix

open (TEXT,"$textfile") || die "Could not open $textfile";

while ($line = <TEXT>){
	chomp($line);
	
	(my $class1, my $class2, my $class3, my $class4, my $class5, my $class6, my $class7, my $class8, my $class9, my$class10, my $class11, my $class12, my $class13, my $class14, my $class15, my $class16, my $class17, my $class18, my $class19, my $class20, my $class21, my $class22 ) = split(/ /,$line);
	
		$w[$i][0] = $class1;
		$w[$i][1] = $class2;
		$w[$i][2] = $class3;
		$w[$i][3] = $class4;
		$w[$i][4] = $class5;
		$w[$i][5] = $class6;
		$w[$i][6] = $class7;
		$w[$i][7] = $class8;
		$w[$i][8] = $class9;
		$w[$i][9] = $class10;
		$w[$i][10] = $class11;
		$w[$i][11] = $class12;
		$w[$i][12] = $class13;
		$w[$i][13] = $class14;
		$w[$i][14] = $class15;
		$w[$i][15] = $class16;
		$w[$i][16] = $class17;
		$w[$i][17] = $class18;
		$w[$i][18] = $class19;
		$w[$i][19] = $class20;
		$w[$i][20] = $class21;
		$w[$i][21] = $class22;
		
		$i = $i +1;
	
}

close(TEXT);

#-------------------------------------------------
#-------Defining gradients and stepfactors--------
#-------------------------------------------------
		
#my @w;			
my @wo;		
my @dw;			
my @dwo;
my @sw;
my $sinit = 10;

for (my $i = 0; $i <= $Nclass_birds; $i++){
	for (my $j = 0; $j < $Nclass_birds; $j++){
		#$w[$j][$i] = 0.01 * (rand() - 0.5);
		$dw[$j][$i] = 0;
		$dwo[$j][$i] = 0;
		$wo[$j][$i] = 0;
		$sw[$j][$i] = 1 * $sinit;
	}
}


#---------------------------------------------
#--------Defining various variables-----------
#---------------------------------------------

my $ud = 0.15;
my $alpha = 0.5 / ($counter1 + 1);
my $Ed_old = -1000; 

#--------------------------------------------
#-------------Totiter itrations--------------
#--------------------------------------------

my $iter = 0;
my $rd_old = -1;
my @cdev1;
my @cdev2;
my @cdev3;
my @zd;
my @bd;
my @bd_sum;

while ($iter <= $totiter){
	
	#########Initializing misclassification matrix
	for (my $i = 0; $i < $Nclass_birds; $i++){
		for (my $j = 0; $j < $Nclass_birds; $j++){
			$cdev1[$i][$j] = 0;
			$cdev2[$i][$j] = 0;
			$cdev3[$i][$j] = 0;
		}
	}
	
	#####reseting derivatives
	for (my $i = 0; $i <= $Nclass_birds; $i++){
		for (my $j = 0; $j < $Nclass_birds; $j++){
			$dw[$j][$i] = 0;
			$dwo[$j][$i] = 0;
			#$wo[$j][$i] = 0;
		}
	}
	
	$iter = $iter + 1; #increasing counter
	
	########Forward calculations
	@zd = @{ matrix_multiply(\@w, \@frame_distr_norm) };
	
	my @sum;
	my @bd_norm;
	
	for (my $i = 0; $i <= $counter1; $i++){
		$bd_sum[$i] = 0;
		for (my $j = 0; $j < $Nclass_birds; $j++){
			$bd[$j][$i] = exp($zd[$j][$i]);
			$bd_sum[$i] = $bd_sum[$i] + $bd[$j][$i];
		}
	}
	
	for (my $i = 0; $i <= $counter1; $i++){
		for (my $j = 0; $j < $Nclass_birds; $j++){
			$bd_norm[$j][$i] = $bd[$j][$i] / $bd_sum[$i];
		}
	}
	
	#####Training criteria
	my @tb;
	my $Ed = 0;    
	
	for (my $i = 0; $i <= $counter1; $i++){
		for (my $j = 0; $j < $Nclass_birds; $j++){
			$tb[$j][$i] = $Ttrain[$j][$i] * log($bd_norm[$j][$i]);
			$Ed = $Ed + $tb[$j][$i];
		}
	}
	
	$Ed = (100 * $Ed) / ($counter1 + 1);
	#print "$Ed, ";
	
	#####Sorting bd matrix
	my @bd_norm_sorted;
	my @bd_norm_sorted_temp;

	for (my $i = 0; $i <= $counter1; $i++){
		@bd_norm_sorted_temp = sort {$b->[$i] <=> $a->[$i]} @bd_norm;
		for (my $j = 0; $j < $Nclass_birds; $j++){
			$bd_norm_sorted[$j][$i] = $bd_norm_sorted_temp[$j][$i];
		}
	}
	
	######Finding indexes of the sorted matrix
	my @file_rec;
	
	for (my $i = 0; $i <= $counter1; $i++){
		$file_rec[0][$i] = 0;
		$file_rec[1][$i] = 0;
		$file_rec[2][$i] = 0;
		for (my $j = 0; $j < $Nclass_birds; $j++){
			if ($bd_norm_sorted[0][$i] == $bd_norm[$j][$i]){				
				$file_rec[0][$i] = $j;
			}
			if ($bd_norm_sorted[1][$i] == $bd_norm[$j][$i]){		
				$file_rec[1][$i] = $j;
			}
			if ($bd_norm_sorted[2][$i] == $bd_norm[$j][$i]){				
				$file_rec[2][$i] = $j;
			} 
		}
	}
	
	
	
	###########Calculating misclassification matrices
	
	my @classdev;
	my @cdev;
	
	for (my $i = 0; $i <= $counter1; $i++){
		$classdev[0][$i] = $file_rec[0][$i];
		$classdev[1][$i] = $file_rec[1][$i];
		$classdev[2][$i] = $file_rec[2][$i];
	}
	
	my $count_file_class1 = 0;
	my $count_file_class2 = 0;
	my $count_file_class3 = 0;
	
	##Initializing
	for (my $j = 0; $j < $Nclass_birds; $j++){
		for (my $k = 0; $k < $Nclass_birds; $k++){
			$cdev[$j][$k] = 0;
		}
	}
	
	##Mis matrix 
	
	for (my $i = 0; $i <= $counter1; $i++){
		$count_file_class1 = 0;
		for (my $j = 0; $j < $Nclass_birds; $j++){
			for (my $k = 0; $k < $Nclass_birds; $k++){
				if ($file_class[$i] == $j && $file_rec[0][$i] == $k){
					$count_file_class1 +=1;
					$cdev[$j][$k] = $cdev[$j][$k] + $count_file_class1;
					#print $cdev[$j][$k]; 
				}
				#print "\n";
			}
		}
	}
	
	####Number of correct classifications 1best, 2best and 3best
	
	$count_file_class1 = 0;
	$count_file_class2 = 0;
	$count_file_class3 = 0;
	
	for (my $i = 0; $i <= $counter1; $i++){
		if ($file_rec[0][$i] == $file_class[$i]){
			$count_file_class1 +=1;
		}
		if ($file_rec[1][$i] == $file_class[$i]){
			$count_file_class2 +=1;
		}
		if ($file_rec[2][$i] == $file_class[$i]){
			$count_file_class3 +=1;
		}
	}
	
	my $numcorr1d = $count_file_class1;
	my $numcorr2d = $count_file_class1 + $count_file_class2;
	my $numcorr3d = $count_file_class1 + $count_file_class2 + $count_file_class3; 
	
	my @dzd;
				
	$perr1[$set] = 100 * (1 - ($numcorr1d / ($counter1 + 1)));
	$perr2[$set] = 100 * (1 - ($numcorr2d / ($counter1 + 1)));
	$perr3[$set] = 100 * (1 - ($numcorr3d / ($counter1 + 1)));			
	
	print "$perr1[$set], $perr2[$set], $perr3[$set]\n";
	
	if ($Ed_old == $Ed_old){
		
		##Hide the old parameters
		@wo = @w;
		@dwo = @dw;
		$Ed_old = $Ed;
		
		##Bakward loop of error and calculation of delta-parameters and updating step factor
		for (my $i = 0; $i <= $counter1; $i++){
			for (my $j = 0; $j < $Nclass_birds; $j++){
				$dzd[$j][$i] = -($bd_norm[$j][$i] - $Ttrain[$j][$i]); ###skal kanskje bruke bd_sorted
				$dzd[$j][$i] = $dzd[$j][$i] * $alpha;
			}
		}
		
		my @frame_distr_norm_trans = pivot(@frame_distr_norm);
		
		@dw = @{ matrix_multiply(\@dzd, \@frame_distr_norm_trans) };
		
		for (my $i = 0; $i <= $Nclass_birds; $i++){
			for (my $j = 0; $j < $Nclass_birds; $j++){
				$dw[$j][$i] = $dw[$j][$i] * $sw[$j][$i];
				$sw[$j][$i] = $sw[$j][$i] + ($ud * ($sw[$j][$i] * sign($dwo[$j][$i], $dw[$j][$i])));
				$w[$j][$i] = (0.999 * $w[$j][$i]) + $dw[$j][$i];
			}
		}
	}
}


$Best_1 = $Best_1 + $perr1[$set]/5;
$Best_2 = $Best_2 + $perr2[$set]/5;
$Best_3 = $Best_3 + $perr3[$set]/5;

open(my $fh, '>', "classificationMatrix$set.txt");
for (my $i = 0; $i < $Nclass_birds; $i++){
	for (my $j = 0; $j <= $Nclass_birds; $j++){
		print $fh $w[$i][$j];
		print $fh " ";
	}
	print $fh "\n";
}
close $fh;
print "done\n";

}


print "$Best_1    $Best_2     $Best_3 \n";