#!/usr/bin/perl


use Getopt::Std;

# Directory structure for training
$d="./config";                      # input data files, part of installation
$w="./work";                        # directory for generated index files, 
                                    # edit scripts, etc.
$t="./tools";                       # sub-scripts, part of installation
$m="./models";                      # output HMM root directory
$f="features";          # feature file root directory


# Various files needed 
$PhoneMlf="lists/Kvirrevitt_new.mlf";  # Phone level MLF from initial segm.
#$PhoneMlf="MLF/train5adaption5_threshold.mlf"; 
$list="lists/train5.scp";           # database information file
#$Proto="lists/proto_s1_m1_dc.pcf";  # Specifications for the HMM prototype
$Proto="lists/proto.pcf";  # Specifications for the HMM prototype
# $Hcomp_conf="$d/hcompv_mfc.cfg";
# $Hrest_conf="$d/herest_mfc.cfg";
$Hcomp_conf="$d/hcompv_mfc.cfg";
$Hrest_conf="$d/herest_mfc.cfg";

# Parameters for training 
$nmix=1;                            # Number of mix components for initialization
$maxmix=32;                         # Max number of mixture components
$moniter1=20;                        # Iterations of HInit and HRest (1 mix)
$moniter2=40;                        # Iterations of HRest (> 1 Gaussian mix)


# Read initial MLF and directory to put models (optional)

getopts('p:');
$partition=2;
if ($opt_p){$partition=$opt_p;}

# Check/create subdirectories...

foreach $tmp ($t,$d) {die "Directory $tmp not found\n" unless (-d $tmp);}
$ENV{'PATH'} = "./$t".":".$ENV{'PATH'};         # put tools directory first in path
require("$t/MMF_subs.pl");                      # load common subroutine libraries
require("$t/common.pl");




$list="lists/train"."$partition".".scp";
$m="$m"."$partition";
$w="$w"."$partition";
$Dict="$w/Kvirrevitt.dict";      # Dictionary
$Phones="$w/Kvirrevitt.lis";     # Class inventory
$Trainset="$w/trainset.scp";        # List of training files

# Then we are ready to start processing...

print "Trainmodels_Bootstrap started ",`date`,"\n";
goto skip;       # restart the script by moving the skip label to the
                 # point where it stopped

skip:            # start processing again here
foreach $tmp ($m,$w){&makedir($tmp)};

$cmd="cp $list $Trainset";
&run("$cmd");


open(PHL,"$PhoneMlf") || die "Could not open $PhoneMlf\n";
%phlist=();
while(<PHL>){
    chomp;
    if ($_ =~ /^[\"\.\#]/) {next;}
    ($start,$end,$phone)=split(' ');
    $phone=~ s/^\s+//;
    $phone=~ s/\s+$//;
    $phlist{$phone}=$phlist{$phone}+1;
}
close(PHL);

open(PHN,">$Phones") || die "Could not open $Phones\n";
open(DIC,">$Dict") || die "Could not open $Dict\n";
foreach $phone (sort keys %phlist) {
    if ($phone eq "sil"){next;}
    print PHN "$phone\n";
    print DIC "$phone $phone\n";
}
close(PHN);
close(DIC);


$startDir=0;

print "HInit/HCompV based on initial segmentation\n";
#----------------------------------------------------------

$srcdir="$m/state1_$startDir";
$tgtdir="$m/state1_$nmix";
&makedir($srcdir);
&makedir($tgtdir);
open(PRTIN,"$Proto") || die "Could not open HMM prototype spec file, $Proto for reading\n";
open(PRTOUT,">$w/proto.pcf") || die "Could not open HMM prototype spec file, $Proto for writing\n";
while(<PRTIN>){
	chomp;
	if ($_ =~ /^outDir/){
		print PRTOUT "outDir: $srcdir\n";
	} else {
		print PRTOUT "$_\n";
	}
}
close(PRTIN);
close(PRTOUT);
#$cmd="cat $Proto | sed 's/proto\/$srcdir2\/' > $w/proto.pcf";
#print "$cmd\n";
#&run("$cmd");
$cmd="$t/MakeProtoHMMSet $w/proto.pcf";
&run("$cmd");
#&MkMacro($ProtoMMF,"$srcdir/$Macro");

# Initialize with a) global variance, b) <20 iterations of HInit
# For all phones in $Phones:
open(PHONES,"$Phones") || die "Could not open $Phones\n";
while(<PHONES>){
	chomp;
	$Ph=$_;
	$cmd="HCompV -A -C $Hcomp_conf -M $srcdir -S $Trainset $srcdir/$Ph >> dump";
	&run("$cmd");
}
close(PHONES);

print "Baum-Welch reestimation using HRest\n";
#----------------------------------------------------------- 

$tgtdir="$m/state1_$nmix";
&makedir($tgtdir);

# For all phones in $Phones:

open(PHONES,"$Phones") || die "Could not open $Phones\n";
while(<PHONES>){
	chomp;

	&run("HRest -A -T 1 -w 2.0 -i $moniter1 -I $PhoneMlf -l $_ -t " .
             "-C $Hrest_conf -H $srcdir/$_ -M $tgtdir -S $Trainset " .
	     "$tgtdir/$_ >> dump");
}
close(PHONES);
$tgthmm="$tgtdir.mmf";
&MkMMF2($tgthmm,$Phones,$tgtdir,A);


print "Training mixture monophones ".`date`;
#-------------------------------------------

sub monomixup {
    local($nmix,$sourcehmm,$targethmm,$phone) = @_;
    &printfile("$w/mix${nmix}.hed", "MU $nmix {*.state[2-4].mix}");
    # print " ...updating to $nmix mixtures in $targethmm\n";
    &run("HHEd -A -H $sourcehmm -w $targethmm $w/mix${nmix}.hed ".
	  "Modelnames/$phone");
    # print " ...finished ".`date`;
}

$mmix=$nmix;
$sourcedir="$m/state1_${nmix}";

while ($mmix < $maxmix) {
    
    $newmix=2*$mmix;
    $targetdir="$m/state1_${newmix}";
	&makedir($targetdir);
    $mmix=$newmix;
    print "training $targethmm for $newmix mixtures\n";
	open(PHONES,"$Phones") || die "Could not open $Phones\n";
	while(<PHONES>){
		chomp;
		$targethmm="$targetdir/$_";
		$sourcehmm="$sourcedir/$_";
#		&run("echo $_ > $w/temp");
    	&monomixup($newmix,$sourcehmm,$targethmm,$_);
#		&run("rm -f $w/temp");
		&run("HRest -A -T 1 -i $moniter2 -I $PhoneMlf -l $_ -t " .
	             "-C $Hrest_conf -H $targethmm -S $Trainset " .
		     "-M $targetdir $targethmm >> dump");
	}
	close(PHONES);
	$tgthmm="$targetdir.mmf";
	&MkMMF2($tgthmm,$Phones,$targetdir,"A");
	$sourcedir=$targetdir;
}

print "TrainModels finished ",`date`;
